import { Camera } from "../../libraries/Camera.js";
import { Character } from "../../libraries/components/Character.js";
import { TileMap } from "../../libraries/components/TileMap.js";
import { Point } from "../../libraries/spatial/Point.js";
import { TweenManager, Tween, Easing } from "../../libraries/Tween.js";

export class OverworldState {
    constructor(view) {
        this.stateMachine = view.stateMachine
        this.canvasBounds = null
        this.camera = new Camera()
        this.camera.scale = new Point(2, 2)
        this.tweenManager = new TweenManager()
        this.characterAtlas = view.characterAtlas

        this.playerPosition = new Point(0,0);
        this.keysPressed = {}
        this.player = new Character(this.characterAtlas, 0, 0)
        this.map = new TileMap()
        this.keypressDelay = 0
    }

    init(self) {
        this.player.load()
        this.putPlayerOnMap(this.player.room)
    }

    async putPlayerOnMap(room) {
        await this.map.loadFromApi(this.player.room)
        if (!this.map.getTile(this.map.mapData.exits[0][0], this.map.mapData.exits[0][1])) {
            this.player.room = "room5"
            this.player.save()
            await this.putPlayerOnMap("room5")
        }
        this.player.position.x = this.map.mapData.exits[0][0] * 16
        this.player.position.y = this.map.mapData.exits[0][1] * 16
        this.camera.position.x = this.camera.scale.x * this.player.position.x
        this.camera.position.y = this.camera.scale.y * this.player.position.y
    }

    draw(ctx, scaledCanvas) {
        this.canvasBounds = scaledCanvas.bounds;
        this.camera.draw(ctx, scaledCanvas, () => {
            this.map.draw(ctx, scaledCanvas)
            this.map.drawDecals(ctx, scaledCanvas)
            this.player.draw(ctx, scaledCanvas)
            this.map.drawOverDecals(ctx, scaledCanvas)
        })
    }

    update(delta) {
        let gridPosition = new Point(Math.floor(this.player.position.x / 16), Math.floor(this.player.position.y / 16))

        let currentTime = new Date().getTime()
        if(this.keypressDelay <= currentTime) { 
            if(this.keysPressed["w"] || this.keysPressed["ArrowUp"]){
                gridPosition.y--
                this.keypressDelay = currentTime + 150
            }
            if(this.keysPressed["a"] || this.keysPressed["ArrowLeft"]){
                gridPosition.x--
                this.keypressDelay = currentTime + 150
            }
            if(this.keysPressed["s"] || this.keysPressed["ArrowDown"]){
                gridPosition.y++
                this.keypressDelay = currentTime + 150
            }
            if(this.keysPressed["d"] || this.keysPressed["ArrowRight"]){
                gridPosition.x++
                this.keypressDelay = currentTime + 150
            }
        }

        this.mapMovementCalculations(gridPosition)

        this.camera.update(delta);
        this.tweenManager.update()
    }

    enter() {
        this.init(this.stateMachine.getState("view"))
        this.registeredEvents = {}
        this.registeredEvents["resize"] = this.onResize
        this.registeredEvents["keydown"] = this.onKeyDown
        this.registeredEvents["keyup"] = this.onKeyUp
        this.registeredEvents["touchend"] = this.onTouchEnd

        for(let index in this.registeredEvents) {
            window.addEventListener(index, this.registeredEvents[index].bind(this))
        }
    }
    leave() {
        for(let index in this.registeredEvents) {
            window.removeEventListener(index, this.registeredEvents[index])
        }
        this.tweenManager.clear()
    }

    onFinish() {
    }

    onResize() {
    }

    onKeyDown(event) {
        this.keysPressed[event.key] = true
    }

    onKeyUp(event) {
        this.tileMoved = false
        this.keysPressed[event.key] = false
        switch(event.key) {
            case "Escape":
                this.stateMachine.transitionTo("charactercreation")
                break
        }
    }

    onTouchEnd(event) {
        let touchPosition = new Point(event.changedTouches[0].clientX, event.changedTouches[0].clientY)
        let worldTouchPosition = this.camera.screenToWorld(touchPosition)
        let delta = worldTouchPosition.difference(this.player.position)

        let gridPosition = new Point(Math.floor(this.player.position.x / 16), Math.floor(this.player.position.y / 16))
        let currentTime = new Date().getTime()
        if(Math.abs(delta.x) > Math.abs(delta.y)) {
            if(Math.sign(delta.x) < 0) {
                gridPosition.x--
            } else {
                gridPosition.x++
            }
        } else if(Math.abs(delta.x) < Math.abs(delta.y)) {
            if(Math.sign(delta.y) < 0) {
                gridPosition.y--
            } else {
                gridPosition.y++
            }
        }
        this.keypressDelay = currentTime + 150

        this.mapMovementCalculations(gridPosition)
    }

    mapMovementCalculations(gridPosition) {
        this.map.updateMovement(gridPosition, (gridPosition, room) => {
            let shouldSave = false
            if (room != this.player.room) {
                shouldSave = true
            }
            this.player.room = room
            this.player.position.x = 16 * gridPosition.x
            this.player.position.y = 16 * gridPosition.y
            this.camera.position.x = this.camera.scale.x * this.player.position.x
            this.camera.position.y = this.camera.scale.y * this.player.position.y
            if (shouldSave) {
                if (!this.map.getTile(gridPosition.x, gridPosition.y)) {
                    alert("This room is incomplete! Enter the editor for more information! (Shift + e)")
                }
                this.player.save()
            }
        })
    }
}
