import { Camera } from "../../libraries/Camera.js";
import { Point } from "../../libraries/spatial/Point.js";
import { SpritesheetAtlas } from "../../libraries/SpritesheetAtlas.js";
import { TweenManager, Tween, Easing } from "../../libraries/Tween.js";

export class SheetTileSelectorState {
    constructor(view) {
        this.stateMachine = view.stateMachine
        this.canvasBounds = null
        this.camera = new Camera()
        this.camera.scale = new Point(2, 2)
        this.tweenManager = new TweenManager()

        this.keysPressed = {}
        this.primaryTile = new Point(0,0)
        this.secondaryTile = new Point(1,0)
        this.hoveredTile = new Point(2,0)
    }

    init(self) {
        let params = new URLSearchParams(window.location.search)
        let sheet = params.get('sheet');
        this.atlas = new SpritesheetAtlas()
        this.atlas.load(sheet, {width: 16, height: 16, margin: 1})
        this.camera.position.x = 300
        this.camera.position.y = 200

        let tileString = localStorage.getItem('teetopia-editor-primary') ?? "0,0"
        let tile = tileString.split(",")
        this.primaryTile.x = tile[0] * 17
        this.primaryTile.y = tile[1] * 17

        let secondTileString = localStorage.getItem('teetopia-editor-secondary') ?? "1,0"
        let secondTile = secondTileString.split(",")
        this.secondaryTile.x = secondTile[0] * 17
        this.secondaryTile.y = secondTile[1] * 17
    }

    draw(ctx, scaledCanvas) {
        this.canvasBounds = scaledCanvas.bounds;

        if(!this.atlas.loaded) {
            return
        }

        this.camera.draw(ctx, scaledCanvas, (ctx, scaledCanvas) => {
            ctx.drawImage(this.atlas.spritesheetImage, 0, 0)

            ctx.strokeStyle = "green"
            ctx.lineWidth = 2
            ctx.beginPath();
            ctx.rect(
                this.primaryTile.x,
                this.primaryTile.y,
                16,
                16,
            )
            ctx.stroke();

            ctx.strokeStyle = "blue"
            ctx.lineWidth = 2
            ctx.beginPath();
            ctx.rect(
                this.secondaryTile.x,
                this.secondaryTile.y,
                16,
                16,
            )
            ctx.stroke();

            ctx.strokeStyle = "rgba(255,255,255,0.5)"
            ctx.lineWidth = 2
            ctx.beginPath();
            ctx.rect(
                this.hoveredTile.x,
                this.hoveredTile.y,
                16,
                16,
            )
            ctx.stroke();

            ctx.textAlign = "right"
            ctx.font = "6px Arial"
            ctx.fillStyle =  "rgba(255,255,255,0.9)"
            ctx.fillText(`${this.hoveredTile.x / 17},${this.hoveredTile.y / 17}`, this.hoveredTile.x + 16, this.hoveredTile.y + 16)

        })
    }

    update(delta) {
        if(this.keysPressed['w'] || this.keysPressed['ArrowUp']) {
            this.camera.position.y -= 8
        }
        if(this.keysPressed['a'] || this.keysPressed['ArrowLeft']) {
            this.camera.position.x -= 8
        }
        if(this.keysPressed['s'] || this.keysPressed['ArrowDown']) {
            this.camera.position.y += 8
        }
        if(this.keysPressed['d'] || this.keysPressed['ArrowRight']) {
            this.camera.position.x += 8
        }

        this.camera.update(delta);
        this.tweenManager.update()
    }

    enter() {
        this.init(this.stateMachine.getState("view"))
        this.registeredEvents = {}
        this.registeredEvents["resize"] = this.onResize
        this.registeredEvents["keydown"] = this.onKeyDown
        this.registeredEvents["keyup"] = this.onKeyUp
        this.registeredEvents["mousemove"] = this.onMouseMove
        this.registeredEvents['click'] = this.onClick
        this.registeredEvents['contextmenu'] = this.onClick

        for(let index in this.registeredEvents) {
            window.addEventListener(index, this.registeredEvents[index].bind(this))
        }
    }
    leave() {
        for(let index in this.registeredEvents) {
            window.removeEventListener(index, this.registeredEvents[index])
        }
        this.tweenManager.clear()
    }

    onFinish() {
    }

    onResize() {
    }

    onKeyDown(event) {
        this.keysPressed[event.key] = true
    }

    onKeyUp(event) {
        this.keysPressed[event.key] = false
        // switch(event.key) {
        //     default:
        // }
    }

    onMouseMove(event) {
        let clickPosition = new Point(event.clientX, event.clientY)
        let worldClickPosition = this.camera.screenToWorld(clickPosition)
        worldClickPosition.snapToSpritesheetGrid(this.atlas.tileDimensions)
        this.hoveredTile.x = worldClickPosition.x
        this.hoveredTile.y = worldClickPosition.y
    }

    onClick(event) {
        event.preventDefault()
        let clickPosition = new Point(event.clientX, event.clientY)
        let worldClickPosition = this.camera.screenToWorld(clickPosition)
        worldClickPosition.snapToSpritesheetGrid(this.atlas.tileDimensions)
        

        if(event.button == 2) {
            this.secondaryTile.x = worldClickPosition.x
            this.secondaryTile.y = worldClickPosition.y
            let secondaryData = `${this.secondaryTile.x / 17},${this.secondaryTile.y / 17},${this.atlas.assetsPath}`
            localStorage.setItem('teetopia-editor-secondary', secondaryData)
        } else {
            this.primaryTile.x = worldClickPosition.x
            this.primaryTile.y = worldClickPosition.y
            let primaryData = `${this.primaryTile.x / 17},${this.primaryTile.y / 17},${this.atlas.assetsPath}`
            localStorage.setItem('teetopia-editor-primary', primaryData)
        }
        
    }
}
