using System;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Configuration;

using gameapi.Library;
using System.Text.Json;
using Microsoft.AspNetCore.Http;
using System.Collections.Generic;

namespace gameapi.Controllers
{
	[ApiController]
	[Route("api/login")]
	[Produces("application/json")]
	public class LoginController : ControllerBase
	{
		private readonly ILogger<LoginController> logger;
		private readonly IConfiguration configuration;

		private JsonFileReader jsonFileReader;
		private JsonFileSaver jsonFileSaver;
		private SHA256HashGenerator sha256HashGenerator;
        private SessionRetriever sessionRetriever;

        public LoginController(IConfiguration config, ILogger<LoginController> log)
		{
			configuration = config;
			logger = log;
			this.jsonFileReader = new JsonFileReader();
			this.jsonFileSaver = new JsonFileSaver();
			this.sha256HashGenerator = new SHA256HashGenerator();
			this.sessionRetriever = new SessionRetriever();
		}

		[HttpPost()]
		public IActionResult Post([FromBody] dynamic body)
		{
			var emailProp = body.GetProperty("username");
			var email = emailProp.ToString();
			var passProp = body.GetProperty("password");
			var password = passProp.ToString();

			List<User> users = this.GetUsers();
			User existingUser = users.Find(x => x.email == email && x.password == this.GetPasswordHash(password));
			if (existingUser == null)
			{
				logger.LogInformation($"Login failed: {email}");
				dynamic errorResponse = new
				{
					responseType = StatusCodes.Status400BadRequest,
					responseMessage = "Username or password is invalid."
				};
				return StatusCode(StatusCodes.Status403Forbidden, errorResponse);
			}

			var token = sha256HashGenerator.Get($"{existingUser.email}{DateTime.Now}");

			var sessionRecord = new
			{
				token = token,
				id = existingUser.id,
				email = existingUser.email,
				validUntil = DateTime.Now.AddHours(4)
			};

			this.jsonFileSaver.SaveGameData($"sessions/{token}.json", sessionRecord);
			var displayExistingUser = JsonSerializer.Serialize(existingUser, new JsonSerializerOptions { WriteIndented = true });
			logger.LogInformation($"User signed in: {displayExistingUser}");

			dynamic response = new
			{
				responseType = StatusCodes.Status200OK,
				responseMessage = "Login successful",
				token = token,
				id = existingUser.id,
				email = existingUser.email,
				validUntil = DateTime.Now.AddHours(4)
			};
			return StatusCode(200, response);
		}

		private string GetPasswordHash(string password)
		{
			var salt = GlobalConfiguration.Get().GetSection("server").GetValue<string>("PasswordSalt");
			return sha256HashGenerator.Get($"{password}{salt}");
		}

		private List<User> GetUsers()
		{
			JsonElement usersProp = this.jsonFileReader.ReadGameData("users.json");
			List<User> users = new List<User>();
			foreach (JsonElement userElement in usersProp.EnumerateArray())
			{
				User userToAdd = gameapi.Library.User.Parse(userElement);
				users.Add(userToAdd);
			}
			return users;
		}

		[HttpPost("register")]
		public IActionResult Register([FromBody] dynamic body)
		{
			var emailProp = body.GetProperty("email");
			var email = emailProp.ToString();
			var passProp = body.GetProperty("password");
			var password = passProp.ToString();
			var displayNameProp = body.GetProperty("displayname");
			var displayName = displayNameProp.ToString();

			List<User> users = this.GetUsers();
			User existingUser = users.Find(x => x.email == email);
			if (existingUser != null)
			{
				var displayExistingUser = JsonSerializer.Serialize(existingUser, new JsonSerializerOptions { WriteIndented = true });
				logger.LogInformation($"Attempted registration of existing user: {displayExistingUser}");
				dynamic errorResponse = new
				{
					responseType = StatusCodes.Status400BadRequest,
					responseMessage = "This email is already claimed."
				};
				return StatusCode(StatusCodes.Status400BadRequest, errorResponse);
			}
			var permissions = new List<string>();
			permissions.Add("user");

			var newUser = new User(
				this.GetNewUserId(),
				displayName,
				this.GetPasswordHash(password),
				email,
				permissions,
				"1"
			);

			users.Add(newUser);
			this.jsonFileSaver.SaveGameData("users.json", users);
			logger.LogInformation($"Registering new user: {newUser}");

			dynamic response = new
			{
				responseType = StatusCodes.Status200OK,
				responseMessage = "Registration success!"
			};
			return StatusCode(200, response);
		}

		private string GetNewUserId() {
			return this.sha256HashGenerator.Get($"{DateTime.Now}");
		}

		[HttpPost("session")]
		public IActionResult Session([FromBody] dynamic body)
		{
			var tokenProp = body.GetProperty("token");
			var token = tokenProp.ToString();

			try {
				var existingUser = this.sessionRetriever.GetUserByToken(token);
				logger.LogInformation($"User rejoined session {token}: {existingUser.email}");
				dynamic response = new
				{
					responseType = StatusCodes.Status200OK,
					id = existingUser.id,
					username = existingUser.email,
					displayname = existingUser.displayname,
					permissions = existingUser.permissions,
				};
				return StatusCode(response.responseType, response);
				
			} catch (Exception e) {
				logger.LogError(e.Message);
				dynamic response = new
				{
					responseType = StatusCodes.Status400BadRequest,
					responseMessage = "Failed to load session data."
				};
				return StatusCode(response.responseType, response);
			}
		}
	}
}
